%code requires {

    // Many thanks to
    // http://www.thebinaryidiot.com/archives/2010/03/06/flex-bison-cpp-example/

    #include "../knowledgebase.h"
    namespace Prolog { class PLLexer; }
    using namespace Prolog;
}

%skeleton "lalr1.cc"
%defines
%define namespace "Prolog"
%define parser_class_name "PLParser"
%parse-param { Prolog::PLLexer& scanner }
%parse-param { Prolog::KnowledgeBase** result }
%lex-param { Prolog::PLLexer &scanner }
%union
{
    std::string* sval;
    float* fval;
    int* ival;

    Clause* clause;
    KnowledgeBase* knowledgebase;
    Term* term;
    Terms* terms;
    Goals* goals;
};

%code {
    static int yylex(Prolog::PLParser::semantic_type * yylval, Prolog::PLLexer &scanner);
}

%start S
%token END 0 "end of file"

%token OPEN
%token CLOSE
%token OPENBRACKET
%token CLOSEBRACKET
%token PIPE

%token COMMA
%token SEMICOLON
%token PERIOD
%token IF
%token QUERY

%token <fval> FLOAT
%token <ival> INT
%token <sval> IDENTIFIER
%token <sval> VARIABLE

%% // Produktionen

S:
         KNOWLEDGEBASE END                    { *result = $<knowledgebase>1; }
       | KNOWLEDGEBASE GOALS END              { *result = $<knowledgebase>1; (*result)->goals = $<goals>2; }
       ;

KNOWLEDGEBASE:
         KNOWLEDGEBASE CLAUSE                 { $<knowledgebase>$ = $<knowledgebase>1; $<knowledgebase>$->AddClause($<clause>2); }
       |                                      { $<knowledgebase>$ = new KnowledgeBase(); }
       ;

CLAUSE:
         TERM PERIOD                          { $<clause>$ = new Clause($<term>1, new Terms); }
       | TERM IF TERMS PERIOD                 { $<clause>$ = new Clause($<term>1, $<terms>3); }
       ;

TERM:
         VARIABLE                             { $<term>$ = new Variable($<sval>1); }
       | IDENTIFIER                           { $<term>$ = new CompositeTerm($<sval>1, NULL); }
       | IDENTIFIER OPEN CLOSE                { $<term>$ = new CompositeTerm($<sval>1, new Terms); }
       | IDENTIFIER OPEN TERMS CLOSE          { $<term>$ = new CompositeTerm($<sval>1, $<terms>3); }
       ;

TERMS:
         TERMS COMMA TERM                     { $<terms>$ = $<terms>1; $<terms>$->push_back($<term>3); }
       | TERM                                 { $<terms>$ = new Terms; $<terms>$->push_back($<term>1); }
       ;

GOAL:
         QUERY TERMS PERIOD                   { $<terms>$ = $<terms>2; }
       ;

GOALS:
         GOALS GOAL                           { $<goals>$ = $<goals>1; $<goals>$->push_back($<terms>2); }
       | GOAL                                 { $<goals>$ = new std::vector<Terms*>; $<goals>$->push_back($<terms>1); }
       ;

%%

#include "pl_lexer.h"

void Prolog::PLParser::error(const Prolog::PLParser::location_type &loc, const std::string &msg)
{
    std::cerr << "Error: " << msg << std::endl;
}

static int yylex(PLParser::semantic_type * yylval, PLLexer &scanner)
{
    return scanner.yylex(yylval);
}
